import collections


class Dispatcher(collections.MutableMapping):
    def __init__(self, prototype, prefix='do'):
        self.method_map = dict()
        self._setup_dispatcher(prototype, prefix)

    def __getitem__(self, key):
        return self.method_map[key]

    def __setitem__(self, key, value):
        self.method_map[key] = value

    def __delitem__(self, key):
        del self.method_map[key]

    def __len__(self):
        return len(self.method_map)

    def __iter__(self):
        return iter(self.method_map)

    def __repr__(self):
        return repr(self.method_map)

    def _prefix_callables(self, prototype, prefix):
        names = [name for name in dir(prototype) if name.startswith(prefix)]
        attrs = [getattr(prototype, name) for name in names]
        callables = [attr for attr in attrs if callable(attr)]
        return callables

    def _setup_dispatcher(self, prototype, prefix):
        callables = self._prefix_callables(prototype, prefix)

        for method in callables:
            name = method.__name__[2:]
            self[name] = method




from twisted.internet import reactor
from twisted.application.internet import TCPClient, TCPServer
from twisted.internet.endpoints import TCP4ClientEndpoint, TCP4ServerEndpoint

from .request import JRPCRequest
from .response import JRPCResponse
from .protocol import JRPCClientProtocol, JRPCServerProtocol
from .factory import JRPCClientFactory, JRPCServerFactory
from .dispatcher import Dispatcher

def _buildPeerService(host, port, protocol, factory, service):
    factory = factory(host, port=port)
    factory.protocol = protocol
    return service(port, factory)

def JRPCClientService(protocol, host='localhost', port=9000):
    return _buildPeerService(host, port, protocol, JRPCClientFactory, TCPClient)

def JRPCServerService(protocol, host='localhost', port=9000):
    return _buildPeerService(host, port, protocol, JRPCServerFactory, TCPServer)

def JRPCClientEndpoint(protocol, host='localhost', port=9000):
    endpoint = TCP4ClientEndpoint(reactor, host, port)
    factory = JRPCClientFactory(host, port=port)
    factory.protocol = protocol
    return endpoint.connect(factory)

def JRPCServerEndpoint(protocol, host='localhost', port=9000):
    endpoint = TCP4ServerEndpoint(reactor, host, port)
    factory = JRPCClientFactory(host, port=port)
    factory.protocol = protocol
    return endpoint.connect(factory)

__all__ = [
    'JRPCRequest', 'JRPCResponse',
    'JRPCClientProtocol', 'JRPCServerProtocol',
    'JRPCClientFactory', 'JRPCServerFactory',
    'JRPCClient', 'JRPCServer',
    'Dispatcher',
]



import json

from twisted.python.failure import Failure
from twisted.internet.defer import Deferred, maybeDeferred

from .request import JRPCRequest
from .response import JRPCResponse

class JRPCManager(object):
    """
    Handles the dispatching of remote requests to invoke
    local methods and the sending of requests to remote
    peers.
    """

    def __init__(self, protocol, dispatcher):
        # mapping of ids to deferreds
        self.requests = {}
        # reference back to parent protocol
        self.protocol = protocol
        # dispatch mapping of local handlers
        self.dispatcher = dispatcher
        # next request id
        self._rid = 0

    def call(self, method, *args, **kwargs):
        """
        Call a remote method and return a deferred which will
        be invoked with either the result or failure.
        """
        d = Deferred()
        self.requests[self._rid] = d
        request = JRPCRequest(method, args, kwargs, self._rid)
        self.protocol.sendMessage(request.json())
        self._rid += 1
        return d

    def invoke(self, method, *args, **kwargs):
        """
        Invoke a remote method but do not expect a result.
        """
        request = JRPCRequest(method, args, kwargs)
        self.protocol.sendMessage(request.json())

    def handle(self, payload):
        """
        Handle an incomming message from the remote peer.
        Message may either be remote reqeusts for local
        method invocation or the result of our own requests.
        """
        if isinstance(payload, bytes):
            payload = payload.decode("utf-8")

        try:
            request = JRPCRequest.from_json(payload)
        except Exception as e:
            pass
        else:
            return self.handle_request(request)

        try:
            response = JRPCResponse.from_json(payload)
        except Exception as e:
            pass
        else:
            return self.handle_response(response)

    def handle_request(self, request):
        """
        Handle an incomming request to call a local method.
        If the request contains an id attach callbacks to
        the resulting deferred object to that it may be
        handed back to the requesting peer.
        """
        try:
            handler = self.dispatcher[request.method]
        except KeyError:
            exception = KeyError("`{}` is not an available method.".format(request.method))
            self.return_error(Failure(exception), request.id)
        else:
            if request.id != None:
                d = maybeDeferred(handler, *request.args, **request.kwargs)
                d.addCallback(self.return_response, request.id)
                d.addErrback(self.return_error, request.id)
            else:
                handler(*request.args, **request.kwargs)

    def return_response(self, result, id):
        """
        Return a response back to the peer.
        """
        response = JRPCResponse(id, result=result)
        self.protocol.sendMessage(response.json())

    def return_error(self, failure, id):
        """
        Return an error back to the peer.
        """
        response = JRPCResponse(id,
                                result=failure.getErrorMessage(),
                                error=str(failure.type.__name__))
        self.protocol.sendMessage(response.json())

    def handle_response(self, response):
        """
        Handle a response to a previously requested remote
        method call.
        """
        d = self.requests.pop(response.id)
        if response.error:
            d.errback(response)
        else:
            d.callback(response)



import json

class JRPCResponse(object):
    """
    Represents the result of an executed RPC method request.
    The id is used to match the result with the original
    request. If the error attribute is not None the result
    will contain the details of the failure.
    """

    def __init__(self, id, result=None, error=None):
        # the id of the corresponding request
        self.id = id
        # the result of the RPC request
        self.result = result
        # the error type if any
        self.error = error

    def data(self):
        '''return dictionary representation of the response'''
        return {
            'id': self.id,
            'result': self.result,
            'error': self.error,
        }

    def json(self):
        '''return json representation of the response'''
        return json.dumps(self.data())

    def __str__(self):
        s = self.result
        if self.error:
            s = "{}:{}".format(self.error, self.result)
        return str(s)

    @classmethod
    def from_json(cls, json_str):
        '''create JRPCResponse from json'''
        obj = json.loads(json_str)
        return cls(obj['id'],
                   obj.get('result'),
                   obj.get('error'))




#!/usr/bin/env python

from setuptools import setup

setup(
    name = "jrpc",
    version = "0.1.1",
    packages = ['jrpc'],
    scripts = ['bin/jrpc'],

    install_requires = [
        "Twisted",
        "autobahn",
    ],

    # metadata for upload to PyPI
    author = "Dustin Lacewell",
    author_email = "dlacewell@gmail.com",
    description = "JSON RPC for Autobahn/Twisted",
    keywords = "twisted autobahn rpc json",
    url = "http://github.com/dustinlacewell/jrpc",
)

